@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2019 The Qt Company Ltd.
 ** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtWidgets module of the Qt Toolkit.
@@ -70,6 +70,7 @@
 #include <qtextformat.h>
 #include <qdatetime.h>
 #include <qapplication.h>
+#include <private/qapplication_p.h>
 #include <limits.h>
 #include <qtexttable.h>
 #include <qvariant.h>
@@ -78,7 +79,11 @@ QT_BEGIN_NAMESPACE
 
 static inline bool shouldEnableInputMethod(QTextEdit *textedit)
 {
+#if defined (Q_OS_ANDROID)
+    return !textedit->isReadOnly() || (textedit->textInteractionFlags() & Qt::TextSelectableByMouse);
+#else
     return !textedit->isReadOnly();
+#endif
 }
 
 class QTextEditControl : public QWidgetTextControl
@@ -116,7 +121,7 @@ public:
 };
 
 QTextEditPrivate::QTextEditPrivate()
-    : control(0),
+    : control(nullptr),
       autoFormatting(QTextEdit::AutoNone), tabChangesFocus(false),
       lineWrap(QTextEdit::WidgetWidth), lineWrapColumnOrWidth(0),
       wordWrap(QTextOption::WrapAtWordBoundaryOrAnywhere), clickCausedFocus(0),
@@ -167,6 +172,9 @@ void QTextEditPrivate::init(const QString &html)
     QObject::connect(control, SIGNAL(copyAvailable(bool)), q, SIGNAL(copyAvailable(bool)));
     QObject::connect(control, SIGNAL(selectionChanged()), q, SIGNAL(selectionChanged()));
     QObject::connect(control, SIGNAL(cursorPositionChanged()), q, SLOT(_q_cursorPositionChanged()));
+#if QT_CONFIG(cursor)
+    QObject::connect(control, SIGNAL(blockMarkerHovered(QTextBlock)), q, SLOT(_q_hoveredBlockWithMarkerChanged(QTextBlock)));
+#endif
 
     QObject::connect(control, SIGNAL(textChanged()), q, SLOT(updateMicroFocus()));
 
@@ -187,6 +195,7 @@ void QTextEditPrivate::init(const QString &html)
     vbar->setSingleStep(20);
 
     viewport->setBackgroundRole(QPalette::Base);
+    q->setMouseTracking(true);
     q->setAcceptDrops(true);
     q->setFocusPolicy(Qt::StrongFocus);
     q->setAttribute(Qt::WA_KeyCompression);
@@ -195,9 +204,6 @@ void QTextEditPrivate::init(const QString &html)
 #ifndef QT_NO_CURSOR
     viewport->setCursor(Qt::IBeamCursor);
 #endif
-#if 0 // Used to be included in Qt4 for Q_WS_WIN
-    setSingleFingerPanEnabled(true);
-#endif
 }
 
 void QTextEditPrivate::_q_repaintContents(const QRectF &contentsRect)
@@ -228,6 +234,23 @@ void QTextEditPrivate::_q_cursorPositionChanged()
 #endif
 }
 
+#if QT_CONFIG(cursor)
+void QTextEditPrivate::_q_hoveredBlockWithMarkerChanged(const QTextBlock &block)
+{
+    Q_Q(QTextEdit);
+    Qt::CursorShape cursor = cursorToRestoreAfterHover;
+    if (block.isValid() && !q->isReadOnly()) {
+        QTextBlockFormat::MarkerType marker = block.blockFormat().marker();
+        if (marker != QTextBlockFormat::MarkerType::NoMarker) {
+            if (viewport->cursor().shape() != Qt::PointingHandCursor)
+                cursorToRestoreAfterHover = viewport->cursor().shape();
+            cursor = Qt::PointingHandCursor;
+        }
+    }
+    viewport->setCursor(cursor);
+}
+#endif
+
 void QTextEditPrivate::pageUpDown(QTextCursor::MoveOperation op, QTextCursor::MoveMode moveMode)
 {
     QTextCursor cursor = control->textCursor();
@@ -251,7 +274,7 @@ void QTextEditPrivate::pageUpDown(QTextCursor::MoveOperation op, QTextCursor::Mo
             vbar->triggerAction(QAbstractSlider::SliderPageStepAdd);
         }
     }
-    control->setTextCursor(cursor);
+    control->setTextCursor(cursor, moveMode == QTextCursor::KeepAnchor);
 }
 
 #if QT_CONFIG(scrollbar)
@@ -366,8 +389,8 @@ void QTextEditPrivate::_q_ensureVisible(const QRectF &_rect)
     \section1 Introduction and Concepts
 
     QTextEdit is an advanced WYSIWYG viewer/editor supporting rich
-    text formatting using HTML-style tags. It is optimized to handle
-    large documents and to respond quickly to user input.
+    text formatting using HTML-style tags, or Markdown format. It is optimized
+    to handle large documents and to respond quickly to user input.
 
     QTextEdit works on paragraphs and characters. A paragraph is a
     formatted string which is word-wrapped to fit into the width of
@@ -381,7 +404,7 @@ void QTextEditPrivate::_q_ensureVisible(const QRectF &_rect)
     QTextEdit can display images, lists and tables. If the text is
     too large to view within the text edit's viewport, scroll bars will
     appear. The text edit can load both plain text and rich text files.
-    Rich text is described using a subset of HTML 4 markup, refer to the
+    Rich text can be described using a subset of HTML 4 markup; refer to the
     \l {Supported HTML Subset} page for more information.
 
     If you just need to display a small piece of rich text use QLabel.
@@ -401,12 +424,19 @@ void QTextEditPrivate::_q_ensureVisible(const QRectF &_rect)
     QTextEdit can display a large HTML subset, including tables and
     images.
 
-    The text is set or replaced using setHtml() which deletes any
+    The text can be set or replaced using \l setHtml() which deletes any
     existing text and replaces it with the text passed in the
     setHtml() call. If you call setHtml() with legacy HTML, and then
     call toHtml(), the text that is returned may have different markup,
     but will render the same. The entire text can be deleted with clear().
 
+    Text can also be set or replaced using \l setMarkdown(), and the same
+    caveats apply: if you then call \l toMarkdown(), the text that is returned
+    may be different, but the meaning is preserved as much as possible.
+    Markdown with some embedded HTML can be parsed, with the same limitations
+    that \l setHtml() has; but \l toMarkdown() only writes "pure" Markdown,
+    without any embedded HTML.
+
     Text itself can be inserted using the QTextCursor class or using the
     convenience functions insertHtml(), insertPlainText(), append() or
     paste(). QTextCursor is also able to insert complex objects like tables
@@ -546,7 +576,7 @@ void QTextEditPrivate::_q_ensureVisible(const QRectF &_rect)
     of the movement keystrokes, for example, \e{Shift+Right}
     will select the character to the right, and \e{Shift+Ctrl+Right} will select the word to the right, etc.
 
-    \sa QTextDocument, QTextCursor, {Application Example},
+    \sa QTextDocument, QTextCursor, {Qt Widgets - Application Example},
         {Syntax Highlighter Example}, {Rich Text Processing}
 */
 
@@ -554,8 +584,9 @@ void QTextEditPrivate::_q_ensureVisible(const QRectF &_rect)
     \property QTextEdit::plainText
     \since 4.3
 
-    This property gets and sets the text editor's contents as plain
-    text. Previous contents are removed and undo/redo history is reset
+    \brief the text editor's contents as plain text.
+
+    Previous contents are removed and undo/redo history is reset
     when the property is set. currentCharFormat() is also reset, unless
     textCursor() is already at the beginning of the document.
 
@@ -572,7 +603,7 @@ void QTextEditPrivate::_q_ensureVisible(const QRectF &_rect)
 
 /*!
     \property QTextEdit::undoRedoEnabled
-    \brief whether undo and redo are enabled
+    \brief whether undo and redo are enabled.
 
     Users are only able to undo or redo actions if this property is
     true, and if there is an action that can be undone (or redone).
@@ -659,7 +690,7 @@ qreal QTextEdit::fontPointSize() const
 QString QTextEdit::fontFamily() const
 {
     Q_D(const QTextEdit);
-    return d->control->textCursor().charFormat().fontFamily();
+    return d->control->textCursor().charFormat().fontFamilies().toStringList().value(0, QString());
 }
 
 /*!
@@ -851,7 +882,7 @@ QTextCursor QTextEdit::textCursor() const
 void QTextEdit::setFontFamily(const QString &fontFamily)
 {
     QTextCharFormat fmt;
-    fmt.setFontFamily(fontFamily);
+    fmt.setFontFamilies({fontFamily});
     mergeCurrentCharFormat(fmt);
 }
 
@@ -1085,11 +1116,11 @@ bool QTextEdit::event(QEvent *e)
         d->sendControlEvent(e);
     }
 #else
-    Q_UNUSED(d)
+    Q_UNUSED(d);
 #endif // QT_NO_CONTEXTMENU
 #ifdef QT_KEYPAD_NAVIGATION
     if (e->type() == QEvent::EnterEditFocus || e->type() == QEvent::LeaveEditFocus) {
-        if (QApplication::keypadNavigationEnabled())
+        if (QApplicationPrivate::keypadNavigationEnabled())
             d->sendControlEvent(e);
     }
 #endif
@@ -1214,6 +1245,57 @@ QString QTextEdit::toHtml() const
 }
 #endif
 
+#if QT_CONFIG(textmarkdownreader) && QT_CONFIG(textmarkdownwriter)
+/*!
+    \property QTextEdit::markdown
+
+    This property provides a Markdown interface to the text of the text edit.
+
+    \c toMarkdown() returns the text of the text edit as "pure" Markdown,
+    without any embedded HTML formatting. Some features that QTextDocument
+    supports (such as the use of specific colors and named fonts) cannot be
+    expressed in "pure" Markdown, and they will be omitted.
+
+    \c setMarkdown() changes the text of the text edit.  Any previous text is
+    removed and the undo/redo history is cleared. The input text is
+    interpreted as rich text in Markdown format.
+
+    Parsing of HTML included in the \a markdown string is handled in the same
+    way as in \l setHtml; however, Markdown formatting inside HTML blocks is
+    not supported.
+
+    Some features of the parser can be enabled or disabled via the \a features
+    argument:
+
+    \value MarkdownNoHTML
+           Any HTML tags in the Markdown text will be discarded
+    \value MarkdownDialectCommonMark
+           The parser supports only the features standardized by CommonMark
+    \value MarkdownDialectGitHub
+           The parser supports the GitHub dialect
+
+    The default is \c MarkdownDialectGitHub.
+
+    \sa plainText, html, QTextDocument::toMarkdown(), QTextDocument::setMarkdown()
+    \since 5.14
+*/
+#endif
+
+#if QT_CONFIG(textmarkdownreader)
+void QTextEdit::setMarkdown(const QString &markdown)
+{
+    Q_D(const QTextEdit);
+    d->control->setMarkdown(markdown);
+}
+#endif
+
+#if QT_CONFIG(textmarkdownwriter)
+QString QTextEdit::toMarkdown(QTextDocument::MarkdownFeatures features) const
+{
+    Q_D(const QTextEdit);
+    return d->control->toMarkdown(features);
+}
+#endif
 
 /*! \reimp
 */
@@ -1224,7 +1306,7 @@ void QTextEdit::keyPressEvent(QKeyEvent *e)
 #ifdef QT_KEYPAD_NAVIGATION
     switch (e->key()) {
         case Qt::Key_Select:
-            if (QApplication::keypadNavigationEnabled()) {
+            if (QApplicationPrivate::keypadNavigationEnabled()) {
                 // code assumes linksaccessible + editable isn't meaningful
                 if (d->control->textInteractionFlags() & Qt::TextEditable) {
                     setEditFocus(!hasEditFocus());
@@ -1245,14 +1327,14 @@ void QTextEdit::keyPressEvent(QKeyEvent *e)
             break;
         case Qt::Key_Back:
         case Qt::Key_No:
-            if (!QApplication::keypadNavigationEnabled()
-                    || (QApplication::keypadNavigationEnabled() && !hasEditFocus())) {
+            if (!QApplicationPrivate::keypadNavigationEnabled()
+                    || (QApplicationPrivate::keypadNavigationEnabled() && !hasEditFocus())) {
                 e->ignore();
                 return;
             }
             break;
         default:
-            if (QApplication::keypadNavigationEnabled()) {
+            if (QApplicationPrivate::keypadNavigationEnabled()) {
                 if (!hasEditFocus() && !(e->modifiers() & Qt::ControlModifier)) {
                     if (e->text()[0].isPrint())
                         setEditFocus(true);
@@ -1341,7 +1423,7 @@ void QTextEdit::keyPressEvent(QKeyEvent *e)
         switch (e->key()) {
             case Qt::Key_Up:
             case Qt::Key_Down:
-                if (QApplication::keypadNavigationEnabled()) {
+                if (QApplicationPrivate::keypadNavigationEnabled()) {
                     // Cursor position didn't change, so we want to leave
                     // these keys to change focus.
                     e->ignore();
@@ -1350,7 +1432,7 @@ void QTextEdit::keyPressEvent(QKeyEvent *e)
                 break;
             case Qt::Key_Back:
                 if (!e->isAutoRepeat()) {
-                    if (QApplication::keypadNavigationEnabled()) {
+                    if (QApplicationPrivate::keypadNavigationEnabled()) {
                         if (document()->isEmpty() || !(d->control->textInteractionFlags() & Qt::TextEditable)) {
                             setEditFocus(false);
                             e->accept();
@@ -1374,9 +1456,11 @@ void QTextEdit::keyPressEvent(QKeyEvent *e)
 */
 void QTextEdit::keyReleaseEvent(QKeyEvent *e)
 {
-#ifdef QT_KEYPAD_NAVIGATION
     Q_D(QTextEdit);
-    if (QApplication::keypadNavigationEnabled()) {
+    if (!isReadOnly())
+        d->handleSoftwareInputPanel();
+#ifdef QT_KEYPAD_NAVIGATION
+    if (QApplicationPrivate::keypadNavigationEnabled()) {
         if (!e->isAutoRepeat() && e->key() == Qt::Key_Back
             && d->deleteAllTimer.isActive()) {
             d->deleteAllTimer.stop();
@@ -1426,7 +1510,7 @@ void QTextEdit::resizeEvent(QResizeEvent *e)
         QVariant alignmentProperty = doc->documentLayout()->property("contentHasAlignment");
 
         if (!doc->pageSize().isNull()
-            && alignmentProperty.type() == QVariant::Bool
+            && alignmentProperty.userType() == QMetaType::Bool
             && !alignmentProperty.toBool()) {
 
             d->_q_adjustScrollbars();
@@ -1471,7 +1555,7 @@ void QTextEditPrivate::relayoutDocument()
         width = lineWrapColumnOrWidth;
     else if (lineWrap == QTextEdit::NoWrap) {
         QVariant alignmentProperty = doc->documentLayout()->property("contentHasAlignment");
-        if (alignmentProperty.type() == QVariant::Bool && !alignmentProperty.toBool()) {
+        if (alignmentProperty.userType() == QMetaType::Bool && !alignmentProperty.toBool()) {
 
             width = 0;
         }
@@ -1588,7 +1672,7 @@ void QTextEdit::mousePressEvent(QMouseEvent *e)
 {
     Q_D(QTextEdit);
 #ifdef QT_KEYPAD_NAVIGATION
-    if (QApplication::keypadNavigationEnabled() && !hasEditFocus())
+    if (QApplicationPrivate::keypadNavigationEnabled() && !hasEditFocus())
         setEditFocus(true);
 #endif
     d->sendControlEvent(e);
@@ -1600,7 +1684,7 @@ void QTextEdit::mouseMoveEvent(QMouseEvent *e)
 {
     Q_D(QTextEdit);
     d->inDrag = false; // paranoia
-    const QPoint pos = e->pos();
+    const QPoint pos = e->position().toPoint();
     d->sendControlEvent(e);
     if (!(e->buttons() & Qt::LeftButton))
         return;
@@ -1623,7 +1707,7 @@ void QTextEdit::mouseReleaseEvent(QMouseEvent *e)
         d->autoScrollTimer.stop();
         ensureCursorVisible();
     }
-    if (!isReadOnly() && rect().contains(e->pos()))
+    if (!isReadOnly() && rect().contains(e->position().toPoint()))
         d->handleSoftwareInputPanel(e->button(), d->clickCausedFocus);
     d->clickCausedFocus = 0;
 }
@@ -1694,7 +1778,7 @@ void QTextEdit::dragLeaveEvent(QDragLeaveEvent *e)
 void QTextEdit::dragMoveEvent(QDragMoveEvent *e)
 {
     Q_D(QTextEdit);
-    d->autoScrollDragPos = e->pos();
+    d->autoScrollDragPos = e->position().toPoint();
     if (!d->autoScrollTimer.isActive())
         d->autoScrollTimer.start(100, this);
     d->sendControlEvent(e);
@@ -1719,7 +1803,7 @@ void QTextEdit::inputMethodEvent(QInputMethodEvent *e)
     Q_D(QTextEdit);
 #ifdef QT_KEYPAD_NAVIGATION
     if (d->control->textInteractionFlags() & Qt::TextEditable
-        && QApplication::keypadNavigationEnabled()
+        && QApplicationPrivate::keypadNavigationEnabled()
         && !hasEditFocus())
         setEditFocus(true);
 #endif
@@ -1754,22 +1838,24 @@ QVariant QTextEdit::inputMethodQuery(Qt::InputMethodQuery query, QVariant argume
         case Qt::ImHints:
         case Qt::ImInputItemClipRectangle:
         return QWidget::inputMethodQuery(query);
+    case Qt::ImReadOnly:
+        return isReadOnly();
     default:
         break;
     }
 
     const QPointF offset(-d->horizontalOffset(), -d->verticalOffset());
-    switch (argument.type()) {
-    case QVariant::RectF:
+    switch (argument.userType()) {
+    case QMetaType::QRectF:
         argument = argument.toRectF().translated(-offset);
         break;
-    case QVariant::PointF:
+    case QMetaType::QPointF:
         argument = argument.toPointF() - offset;
         break;
-    case QVariant::Rect:
+    case QMetaType::QRect:
         argument = argument.toRect().translated(-offset.toPoint());
         break;
-    case QVariant::Point:
+    case QMetaType::QPoint:
         argument = argument.toPoint() - offset;
         break;
     default:
@@ -1777,14 +1863,14 @@ QVariant QTextEdit::inputMethodQuery(Qt::InputMethodQuery query, QVariant argume
     }
 
     const QVariant v = d->control->inputMethodQuery(query, argument);
-    switch (v.type()) {
-    case QVariant::RectF:
+    switch (v.userType()) {
+    case QMetaType::QRectF:
         return v.toRectF().translated(offset);
-    case QVariant::PointF:
+    case QMetaType::QPointF:
         return v.toPointF() + offset;
-    case QVariant::Rect:
+    case QMetaType::QRect:
         return v.toRect().translated(offset.toPoint());
-    case QVariant::Point:
+    case QMetaType::QPoint:
         return v.toPoint() + offset.toPoint();
     default:
         break;
@@ -1837,7 +1923,7 @@ void QTextEdit::changeEvent(QEvent *e)
     if (e->type() == QEvent::ApplicationFontChange
         || e->type() == QEvent::FontChange) {
         d->control->document()->setDefaultFont(font());
-    }  else if(e->type() == QEvent::ActivationChange) {
+    }  else if (e->type() == QEvent::ActivationChange) {
         if (!isActiveWindow())
             d->autoScrollTimer.stop();
     } else if (e->type() == QEvent::EnabledChange) {
@@ -1976,33 +2062,18 @@ void QTextEdit::setOverwriteMode(bool overwrite)
     d->control->setOverwriteMode(overwrite);
 }
 
-#if QT_DEPRECATED_SINCE(5, 10)
-/*!
-    \property QTextEdit::tabStopWidth
-    \brief the tab stop width in pixels
-    \since 4.1
-    \deprecated in Qt 5.10. Use tabStopDistance instead.
-
-    By default, this property contains a value of 80 pixels.
-*/
-
-int QTextEdit::tabStopWidth() const
-{
-    return qRound(tabStopDistance());
-}
-
-void QTextEdit::setTabStopWidth(int width)
-{
-    setTabStopDistance(width);
-}
-#endif
-
 /*!
     \property QTextEdit::tabStopDistance
     \brief the tab stop distance in pixels
     \since 5.10
 
     By default, this property contains a value of 80 pixels.
+
+    Do not set a value less than the \l {QFontMetrics::}{horizontalAdvance()}
+    of the QChar::VisualTabCharacter character, otherwise the tab-character
+    will be drawn incompletely.
+
+    \sa QTextOption::ShowTabsAndSpaces, QTextDocument::defaultTextOption
 */
 
 qreal QTextEdit::tabStopDistance() const
@@ -2187,7 +2258,7 @@ void QTextEdit::setReadOnly(bool ro)
     d->control->setTextInteractionFlags(flags);
     setAttribute(Qt::WA_InputMethodEnabled, shouldEnableInputMethod(this));
     QEvent event(QEvent::ReadOnlyChange);
-    QApplication::sendEvent(this, &event);
+    QCoreApplication::sendEvent(this, &event);
 }
 
 /*!
@@ -2532,27 +2603,6 @@ bool QTextEdit::find(const QString &exp, QTextDocument::FindFlags options)
     return d->control->find(exp, options);
 }
 
-/*!
-    \fn bool QTextEdit::find(const QRegExp &exp, QTextDocument::FindFlags options)
-
-    \since 5.3
-    \overload
-
-    Finds the next occurrence, matching the regular expression, \a exp, using the given
-    \a options. The QTextDocument::FindCaseSensitively option is ignored for this overload,
-    use QRegExp::caseSensitivity instead.
-
-    Returns \c true if a match was found and changes the cursor to select the match;
-    otherwise returns \c false.
-*/
-#ifndef QT_NO_REGEXP
-bool QTextEdit::find(const QRegExp &exp, QTextDocument::FindFlags options)
-{
-    Q_D(QTextEdit);
-    return d->control->find(exp, options);
-}
-#endif
-
 /*!
     \fn bool QTextEdit::find(const QRegularExpression &exp, QTextDocument::FindFlags options)
 
